"""reeln-cli plugin for OpenAI-powered LLM integration (metadata, translation, zoom)."""

from __future__ import annotations

__version__ = "0.8.2"

from reeln_openai_plugin.plugin import OpenAIPlugin

__all__ = ["OpenAIPlugin", "__version__"]
